#ifndef MY_ABC_HERE
#define MY_ABC_HERE
#endif
/* modechange.c -- file mode manipulation

   Copyright (C) 1989-1990, 1997-1999, 2001, 2003-2006, 2009-2015 Free Software
   Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

/* Written by David MacKenzie <djm@ai.mit.edu> */

/* The ASCII mode string is compiled into an array of 'struct
   modechange', which can then be applied to each file to be changed.
   We do this instead of re-parsing the ASCII string for each file
   because the compiled form requires less computation to use; when
   changing the mode of many files, this probably results in a
   performance gain.  */

#include <config.h>

#include "modechange.h"
#include <sys/stat.h>
#include "stat-macros.h"
#include "xalloc.h"
#include <stdlib.h>

#ifdef MY_ABC_HERE
#   include <string.h>
#   include <ctype.h>
#   include <limits.h>
#   include <unistd.h>
#   include <sys/types.h>
#   include <pwd.h>
#   include <grp.h>
#   include <synoacl/synoacl.h>
#   include <synocredentials/synocredentials.h>
#endif

/* The traditional octal values corresponding to each mode bit.  */
#define SUID 04000
#define SGID 02000
#define SVTX 01000
#define RUSR 00400
#define WUSR 00200
#define XUSR 00100
#define RGRP 00040
#define WGRP 00020
#define XGRP 00010
#define ROTH 00004
#define WOTH 00002
#define XOTH 00001
#define ALLM 07777 /* all octal mode bits */

/* Convert OCTAL, which uses one of the traditional octal values, to
   an internal mode_t value.  */
static mode_t
octal_to_mode (unsigned int octal)
{
  /* Help the compiler optimize the usual case where mode_t uses
     the traditional octal representation.  */
  return ((S_ISUID == SUID && S_ISGID == SGID && S_ISVTX == SVTX
           && S_IRUSR == RUSR && S_IWUSR == WUSR && S_IXUSR == XUSR
           && S_IRGRP == RGRP && S_IWGRP == WGRP && S_IXGRP == XGRP
           && S_IROTH == ROTH && S_IWOTH == WOTH && S_IXOTH == XOTH)
          ? octal
          : (mode_t) ((octal & SUID ? S_ISUID : 0)
                      | (octal & SGID ? S_ISGID : 0)
                      | (octal & SVTX ? S_ISVTX : 0)
                      | (octal & RUSR ? S_IRUSR : 0)
                      | (octal & WUSR ? S_IWUSR : 0)
                      | (octal & XUSR ? S_IXUSR : 0)
                      | (octal & RGRP ? S_IRGRP : 0)
                      | (octal & WGRP ? S_IWGRP : 0)
                      | (octal & XGRP ? S_IXGRP : 0)
                      | (octal & ROTH ? S_IROTH : 0)
                      | (octal & WOTH ? S_IWOTH : 0)
                      | (octal & XOTH ? S_IXOTH : 0)));
}

/* Special operations flags.  */
enum
  {
    /* For the sentinel at the end of the mode changes array.  */
    MODE_DONE,

    /* The typical case.  */
    MODE_ORDINARY_CHANGE,

    /* In addition to the typical case, affect the execute bits if at
       least one execute bit is set already, or if the file is a
       directory.  */
    MODE_X_IF_ANY_X,

    /* Instead of the typical case, copy some existing permissions for
       u, g, or o onto the other two.  Which of u, g, or o is copied
       is determined by which bits are set in the 'value' field.  */
    MODE_COPY_EXISTING
  };

/* Description of a mode change.  */
struct mode_change
{
  char op;                      /* One of "=+-".  */
  char flag;                    /* Special operations flag.  */
  mode_t affected;              /* Set for u, g, o, or a.  */
  mode_t value;                 /* Bits to add/remove.  */
  mode_t mentioned;             /* Bits explicitly mentioned.  */
#ifdef MY_ABC_HERE
  SYNO_ACL *acl;
#endif
};

/* Return a mode_change array with the specified "=ddd"-style
   mode change operation, where NEW_MODE is "ddd" and MENTIONED
   contains the bits explicitly mentioned in the mode are MENTIONED.  */

static struct mode_change *
make_node_op_equals (mode_t new_mode, mode_t mentioned)
{
    struct mode_change *p = xmalloc (2 * sizeof *p);
    p->op = '=';
    p->flag = MODE_ORDINARY_CHANGE;
    p->affected = CHMOD_MODE_BITS;
    p->value = new_mode;
    p->mentioned = mentioned;
    p[1].flag = MODE_DONE;
    return p;
}

#ifdef MY_ABC_HERE
static struct mode_change *
make_node_op_equals_acl (mode_t new_mode, SYNO_ACL* acl)
{
  struct mode_change *p = make_node_op_equals(new_mode, CHMOD_MODE_BITS);
  p->acl = acl;
  return p;
}
#endif

/* Return a pointer to an array of file mode change operations created from
   MODE_STRING, an ASCII string that contains either an octal number
   specifying an absolute mode, or symbolic mode change operations with
   the form:
   [ugoa...][[+-=][rwxXstugo...]...][,...]

   Return NULL if 'mode_string' does not contain a valid
   representation of file mode change operations.  */

struct mode_change *
mode_compile (char const *mode_string)
{
  /* The array of mode-change directives to be returned.  */
  struct mode_change *mc;
  size_t used = 0;
  char const *p;

  if ('0' <= *mode_string && *mode_string < '8')
    {
      unsigned int octal_mode = 0;
      mode_t mode;
      mode_t mentioned;

      p = mode_string;
      do
        {
          octal_mode = 8 * octal_mode + *p++ - '0';
          if (ALLM < octal_mode)
            return NULL;
        }
      while ('0' <= *p && *p < '8');

      if (*p)
        return NULL;

      mode = octal_to_mode (octal_mode);
      mentioned = (p - mode_string < 5
                   ? (mode & (S_ISUID | S_ISGID)) | S_ISVTX | S_IRWXUGO
                   : CHMOD_MODE_BITS);
      return make_node_op_equals (mode, mentioned);
    }

  /* Allocate enough space to hold the result.  */
  {
    size_t needed = 1;
    for (p = mode_string; *p; p++)
      needed += (*p == '=' || *p == '+' || *p == '-');
    mc = xnmalloc (needed, sizeof *mc);
  }

  /* One loop iteration for each
     '[ugoa]*([-+=]([rwxXst]*|[ugo]))+|[-+=][0-7]+'.  */
  for (p = mode_string; ; p++)
    {
      /* Which bits in the mode are operated on.  */
      mode_t affected = 0;

      /* Turn on all the bits in 'affected' for each group given.  */
      for (;; p++)
        switch (*p)
          {
          default:
            goto invalid;
          case 'u':
            affected |= S_ISUID | S_IRWXU;
            break;
          case 'g':
            affected |= S_ISGID | S_IRWXG;
            break;
          case 'o':
            affected |= S_ISVTX | S_IRWXO;
            break;
          case 'a':
            affected |= CHMOD_MODE_BITS;
            break;
          case '=': case '+': case '-':
            goto no_more_affected;
          }
    no_more_affected:;

      do
        {
          char op = *p++;
          mode_t value;
          mode_t mentioned = 0;
          char flag = MODE_COPY_EXISTING;
          struct mode_change *change;

          switch (*p)
            {
            case '0': case '1': case '2': case '3':
            case '4': case '5': case '6': case '7':
              {
                unsigned int octal_mode = 0;

                do
                  {
                    octal_mode = 8 * octal_mode + *p++ - '0';
                    if (ALLM < octal_mode)
                      goto invalid;
                  }
                while ('0' <= *p && *p < '8');

                if (affected || (*p && *p != ','))
                  goto invalid;
                affected = mentioned = CHMOD_MODE_BITS;
                value = octal_to_mode (octal_mode);
                flag = MODE_ORDINARY_CHANGE;
                break;
              }

            case 'u':
              /* Set the affected bits to the value of the "u" bits
                 on the same file.  */
              value = S_IRWXU;
              p++;
              break;
            case 'g':
              /* Set the affected bits to the value of the "g" bits
                 on the same file.  */
              value = S_IRWXG;
              p++;
              break;
            case 'o':
              /* Set the affected bits to the value of the "o" bits
                 on the same file.  */
              value = S_IRWXO;
              p++;
              break;

            default:
              value = 0;
              flag = MODE_ORDINARY_CHANGE;

              for (;; p++)
                switch (*p)
                  {
                  case 'r':
                    value |= S_IRUSR | S_IRGRP | S_IROTH;
                    break;
                  case 'w':
                    value |= S_IWUSR | S_IWGRP | S_IWOTH;
                    break;
                  case 'x':
                    value |= S_IXUSR | S_IXGRP | S_IXOTH;
                    break;
                  case 'X':
                    flag = MODE_X_IF_ANY_X;
                    break;
                  case 's':
                    /* Set the setuid/gid bits if 'u' or 'g' is selected.  */
                    value |= S_ISUID | S_ISGID;
                    break;
                  case 't':
                    /* Set the "save text image" bit if 'o' is selected.  */
                    value |= S_ISVTX;
                    break;
                  default:
                    goto no_more_values;
                  }
            no_more_values:;
            }

          change = &mc[used++];
          change->op = op;
          change->flag = flag;
          change->affected = affected;
          change->value = value;
          change->mentioned =
            (mentioned ? mentioned : affected ? affected & value : value);
        }
      while (*p == '=' || *p == '+' || *p == '-');

      if (*p != ',')
        break;
    }

  if (*p == 0)
    {
      mc[used].flag = MODE_DONE;
      return mc;
    }

invalid:
  free (mc);
  return NULL;
}

#ifdef MY_ABC_HERE
SYNO_ACL *get_acl_from_mode_change(struct mode_change* mc)
{
    return mc->acl;
}

SYNO_ACL *trf_mode_to_acl(struct mode_change *mc, const struct stat* file_stats)
{
    SYNO_ACL *acl       = NULL;
    SYNO_ACL_ENTRY *ace = NULL;
    mode_t chng_mode = (mc->affected)?(mc->affected & mc->value): mc->value;
    int ace_ctr = 0;

    ace_ctr += (chng_mode & S_IRWXU)?1:0;
    ace_ctr += (chng_mode & S_IRWXG)?1:0;
    ace_ctr += (chng_mode & S_IRWXO)?1:0;
    acl = SYNOACLAlloc(ace_ctr);
    if ( NULL == acl ) {
        return NULL;
    }

    for(ace=acl->pFirstEnt; ace; ace=ace->pNext){
        ace->inherit = SYNOSDK_ACL_INHERIT_FILE | SYNOSDK_ACL_INHERIT_DIR;
        ace->isAllow = TRUE;
    }

    ace = acl->pFirstEnt;
    if ( chng_mode & S_IRWXU ) {
        if ( DSM_ROOT_UID == file_stats->st_uid ) {
            ace->aid.id = DSM_ADMIN_UID;
        } else {
            ace->aid.id = file_stats->st_uid;
        }
        ace->aid.type = SYNOSDK_ACL_USER;
        ace->perm = SYNOACLModeToPerm(chng_mode >> 6);
        ace = ace->pNext;
    }

    if ( chng_mode & S_IRWXG ) {
        if ( DSM_ROOT_GID == file_stats->st_gid ) {
            ace->aid.id = DSM_ADMIN_GID;
        } else {
            ace->aid.id = file_stats->st_gid;
        }
        ace->aid.type = SYNOSDK_ACL_GROUP;
        ace->perm = SYNOACLModeToPerm(chng_mode >> 3);
        ace = ace->pNext;
    }

    if ( chng_mode & S_IRWXO ) {
        ace->aid.id = SYNOSDK_ACL_UNDEFINED_ID;
        ace->aid.type = SYNOSDK_ACL_GROUP_EVERYONE;
        ace->perm = SYNOACLModeToPerm(chng_mode);
        ace = ace->pNext;
    }

    mc->acl = acl;

    return acl;
}

/**
 * Support more format than SYNOACLAceFromStr
 */
static int
parse_acl_str(char* _acl_str, SYNO_ACL* acl)
{
  SYNO_ACL_ENTRY* ace = SYNOACLAceAlloc();
  char *save_ptr = NULL;
  char *tok = NULL;
  char *c = NULL;
  char *acl_str = calloc(strlen(_acl_str)*2+1, 1);
  int idx = 0;
  int ret = -1;

  if (NULL == ace) {
    goto RET;
  }

  if ( NULL == acl_str ) {
    goto RET;
  }
  // make each '::' become ':*:'
  for(c=_acl_str; *c; c++) {
    acl_str[idx] = *c;
    idx++;
    if( ':' == *c && ':' == *(c+1) ){
      acl_str[idx] = '*';
      idx++;
    }
  }

  // Object type
  tok = strtok_r(acl_str, ":", &save_ptr);
  if(!tok){
      goto RET;
  }
  ace->aid.type = SYNOSDK_ACL_UNKNOWN;
  if (!strcmp(tok, SYNOSDK_ACL_STR_EVERYONE)) {
      ace->aid.type = SYNOSDK_ACL_GROUP_EVERYONE;
  } else if (!strcmp(tok, SYNOSDK_ACL_STR_OWNER)) {
      ace->aid.type = SYNOSDK_ACL_OWNER;
  } else if (!strcmp(tok, SYNOSDK_ACL_STR_USER)) {
      ace->aid.type = SYNOSDK_ACL_USER;
  } else if (!strcmp(tok, SYNOSDK_ACL_STR_GROUP)) {
      ace->aid.type = SYNOSDK_ACL_GROUP;
  } else if (!strcmp(tok, SYNOSDK_ACL_STR_AUTHENTICATEDUSER)) {
      ace->aid.type = SYNOSDK_ACL_AUTHENTICATEDUSER;
  } else if (!strcmp(tok, SYNOSDK_ACL_STR_SYSTEM)) {
      ace->aid.type = SYNOSDK_ACL_SYSTEM;
  } else if (strlen(tok) == 1) {
    switch(*tok) {
      case 'a':
        ace->aid.type = SYNOSDK_ACL_GROUP_EVERYONE;
        break;
      case 'o':
        ace->aid.type = SYNOSDK_ACL_OWNER;
        break;
      case 'u':
        ace->aid.type = SYNOSDK_ACL_USER;
        break;
      case 'g':
        ace->aid.type = SYNOSDK_ACL_GROUP;
        break;
      case 'l':
        ace->aid.type = SYNOSDK_ACL_AUTHENTICATEDUSER;
        break;
      case 's':
        ace->aid.type = SYNOSDK_ACL_SYSTEM;
        break;
      default:
          goto RET;
    }
  } else {
      goto RET;
  }

  // Object name
  tok = strtok_r(NULL, ":", &save_ptr);
  if(!tok){
      goto RET;
  }
  if ( 0 > SYNOACLAceNameToID(ace->aid.type, tok, &(ace->aid.id))) {
      goto RET;
  }

  // allow / deny
  tok = strtok_r(NULL, ":", &save_ptr);
  if(!tok){
      goto RET;
  }
  for(c = tok; *c; c+=1) *c = tolower(*c);
  if ( !strcmp(tok, "allow") ) {
      ace->isAllow = TRUE;
  } else if ( !strcmp(tok, "deny") || strcmp(tok, "*") ) {
      ace->isAllow = FALSE;
  } else {
      goto RET;
  }

  // permission
#define CASE_UNION(C, VAR, VALUE) case C : VAR |= VALUE ; break;

  tok = strtok_r(NULL, ":", &save_ptr);
  if(!tok){
      goto RET;
  }
  ace->perm = 0;
  if (tolower(tok[0]) == 'm') {
    for (c = tok+1; *c; c += 1) {
      switch(tolower(*c)) {
        CASE_UNION('r', ace->perm, SYNOSDK_MODE_READ);
        CASE_UNION('w', ace->perm, SYNOSDK_MODE_WRITE);
        CASE_UNION('x', ace->perm, SYNOSDK_ACL_PERM_EXEC);
        default: goto RET;
      }
    }
  } else if (tolower(tok[0]) == 'f' && !tok[1]) {
      ace->perm |= SYNOSDK_ACL_PERM_FULL_CONTROL;
  } else {
    for (c=tok; *c; c+=1) {
      switch(*c) {
        CASE_UNION('r', ace->perm, SYNOSDK_ACL_PERM_READ);
        CASE_UNION('w', ace->perm, SYNOSDK_ACL_PERM_WRITE);
        CASE_UNION('x', ace->perm, SYNOSDK_ACL_PERM_EXEC);
        CASE_UNION('p', ace->perm, SYNOSDK_ACL_PERM_APPEND);
        CASE_UNION('d', ace->perm, SYNOSDK_ACL_PERM_DEL);
        CASE_UNION('D', ace->perm, SYNOSDK_ACL_PERM_DEL_CHILD);
        CASE_UNION('a', ace->perm, SYNOSDK_ACL_PERM_READ_ATTR);
        CASE_UNION('A', ace->perm, SYNOSDK_ACL_PERM_WRITE_ATTR);
        CASE_UNION('R', ace->perm, SYNOSDK_ACL_PERM_READ_EXT_ATTR);
        CASE_UNION('W', ace->perm, SYNOSDK_ACL_PERM_WRITE_EXT_ATTR);
        CASE_UNION('c', ace->perm, SYNOSDK_ACL_PERM_READ_PERMISSION);
        CASE_UNION('C', ace->perm, SYNOSDK_ACL_PERM_WRITE_PERMISSION);
        CASE_UNION('o', ace->perm, SYNOSDK_ACL_PERM_GET_OWNER_SHIP);
        case '-': continue;
        default: goto RET;
      }
    }
  }

  if( !(ace->perm) ) {
    goto RET;
  }

  // inherit
  tok = strtok_r(NULL, ":", &save_ptr);
  ace->inherit = 0;
  if ( NULL == tok ) {
    ace->inherit = SYNOSDK_ACL_INHERIT_FILE | SYNOSDK_ACL_INHERIT_DIR;
  } else {
    for (c=tok; *c; c+=1) {
      switch(*c) {
        CASE_UNION('f', ace->inherit, SYNOSDK_ACL_INHERIT_FILE);
        CASE_UNION('d', ace->inherit, SYNOSDK_ACL_INHERIT_DIR);
        CASE_UNION('i', ace->inherit, SYNOSDK_ACL_INHERIT_ONLY);
        CASE_UNION('n', ace->inherit, SYNOSDK_ACL_INHERIT_NO_PROPOGATE);
        case '-': break;
        default: goto RET;
      }
    }
  }
#undef CASE_UNION

  if (0 > SYNOACLAceAppend(acl, ace)) {
    goto RET;
  }
  ace = NULL;
  ret = 0;

RET:
  free(ace);
  free(acl_str);
  return ret;
}

struct mode_change *
mode_compile_acl (char *mode_string, bool is_remove_index)
{
  struct mode_change *mc = xmalloc(2 * sizeof(struct mode_change));
  char *acl_str       = NULL;
  char *save_ptr      = NULL;

  if ( !mc ) {
    goto ERROR;
  }

  mc[0].op = mode_string[0]; /*-,+,=*/
  mc[0].flag = MODE_ORDINARY_CHANGE;
  mc[0].affected = CHMOD_MODE_BITS;
  mc[0].mentioned = CHMOD_MODE_BITS;
  mc[0].acl = NULL;
  mc[1].flag = MODE_DONE;

  if ( is_remove_index ) {
    goto RET;
  }

  save_ptr = NULL;

  mc[0].acl = SYNOACLAlloc(0);
  if ( !mc[0].acl ) {
    goto ERROR;
  }
  mode_string += 2; //skip [-+=]a

  acl_str = strtok_r(mode_string, ",", &save_ptr);
  while (acl_str && mc[0].acl->counts < SYNOSDK_ACL_MAX_ACE_COUNT) {
    if ( 0 > parse_acl_str(acl_str, mc[0].acl) ) {
      goto ERROR;
    }

    acl_str = strtok_r(NULL, ",", &save_ptr);
  }

  while(acl_str) acl_str = strtok_r(NULL, ",", &save_ptr);
  goto RET;

ERROR:
  if (mc && mc[0].acl) {
    SYNOACLFree(mc[0].acl);
    mc[0].acl = NULL;
  }

  if (mc) {
    free(mc);
    mc = NULL;
  }

RET:
  return mc;
}

static int cmp(const void *a, const void *b)
{
    return (*(int*)a - *(int*)b);
}

/**
 * if adjust success, nacl merges to oacl and oacl will be return.
 * therefore, there is no need to free any pointer after this function called.
 */
SYNO_ACL *acl_adjust(SYNO_ACL *oacl,
        SYNO_ACL *nacl, int *acl_index, struct mode_change* mc)
{
    SYNO_ACL_ENTRY *ace     = NULL;
    SYNO_ACL_ENTRY *oace    = NULL;
    SYNO_ACL_ENTRY *ace_dup = NULL;
    SYNO_ACE_FILTER acef;
    BOOL isInherited = FALSE;
    int index_num = 0;
    int ctr = 0;

    if (NULL == oacl || (NULL == nacl && NULL == acl_index) || NULL == mc) {
        goto ERROR;
    }

    switch(mc->op)
    {
    case '+':
        for(ace=nacl->pFirstEnt; ace; ace=ace->pNext){
            unsigned int perm = 0;
            unsigned int inherit = ace->inherit;

            ace_dup = SYNOACLAceDup(ace);
            if ( NULL == ace_dup ) {
                goto ERROR;
            }

            for (oace=oacl->pFirstEnt; oace; oace=oace->pNext) {
                if(oace->aid.type == ace_dup->aid.type &&
                        oace->aid.id == ace_dup->aid.id &&
                        oace->inherit == ace_dup->inherit &&
                        oace->isAllow == ace_dup->isAllow) {
                    perm |= (oace->perm);
                }
            }

            memset(&acef, 0, sizeof(acef));
            acef.pIdx = NULL;
            acef.pID = &(ace_dup->aid);
            acef.pPerm = NULL;
            acef.pInheritMode = &(inherit);
            acef.pIsAllow = &(ace_dup->isAllow);
            acef.pIsInherited = &isInherited;

            if ( 0 > SYNOACLAceRemove(oacl, &acef, NULL) ) {
                goto ERROR;
            }

            ace_dup->perm = (perm | ace_dup->perm);
            if ( 0 > SYNOACLAceAppend(oacl, ace_dup) ){
                SYNOACLAceFree(ace_dup);
                goto ERROR;
            }
        }
        break;

    case '-':
        if ( acl_index ) {
            nacl = SYNOACLAlloc(0);
            if ( NULL == nacl ) {
                goto ERROR;
            }
            for (index_num = 0; acl_index[index_num] >= 0; index_num++);
            qsort(acl_index, index_num, sizeof(int), cmp);

            index_num = 0;
            for (oace = oacl->pFirstEnt; oace; oace = oace->pNext) {
                if( acl_index[index_num] < 0 ) {
                    break;
                }
                if( ctr == acl_index[index_num] ) {
                    ace_dup = SYNOACLAceDup(oace);
                    if ( NULL == ace_dup ) {
                        goto ERROR;
                    }
                    if ( 0 > SYNOACLAceAppend(nacl, ace_dup) ) {
                        SYNOACLAceFree(ace_dup);
                        goto ERROR;
                    }
                    index_num++;
                }
                ctr++;
            }
        }

        for (ace=nacl->pFirstEnt; ace; ace=ace->pNext) {
            unsigned int perm = 0;
            unsigned int inherit = ace->inherit;

            for (oace=oacl->pFirstEnt; oace; oace=oace->pNext) {
                if(oace->aid.type == ace->aid.type &&
                        oace->aid.id == ace->aid.id &&
                        oace->inherit == ace->inherit &&
                        oace->isAllow == ace->isAllow) {
                    perm |= oace->perm;
                }
            }
            perm &= ~(ace->perm);

            memset(&acef, 0, sizeof(acef));
            acef.pIdx = NULL;
            acef.pID = &(ace->aid);
            acef.pPerm = NULL;
            acef.pInheritMode = &(inherit);
            acef.pIsAllow = &(ace->isAllow);
            acef.pIsInherited = &isInherited;

            if ( 0 > SYNOACLAceRemove(oacl, &acef, NULL) ) {
                goto ERROR;
            }

            if ( 0 != perm ) {
                ace_dup = SYNOACLAceDup(ace);
                ace_dup->perm = perm;
                if (NULL == ace_dup) {
                    goto ERROR;
                }
                if (0 > SYNOACLAceAppend(oacl, ace_dup)) {
                    SYNOACLAceFree(ace_dup);
                    goto ERROR;
                }
            }
        }

        if (acl_index) {
            SYNOACLFree(nacl);
        }
        break;

    case '=':
        if ( acl_index ) {
            mc->op = '-';
            oacl = acl_adjust(oacl, NULL, acl_index, mc);
            if( NULL == oacl ) {
                goto ERROR;
            }
        } else {
            /**
             * acl_index must be given if syno acl assign
             * however, acl_index is empty if linux perm assign with -k
             * in this case, remove all acl
             */
            SYNOACLFree(oacl);
            oacl = SYNOACLAlloc(0);
            if ( NULL == oacl ) {
                goto ERROR;
            }
        }

        mc->op = '+';
        nacl = acl_adjust(oacl, nacl, NULL, mc);
        if ( NULL == nacl ) {
            SYNOACLFree(oacl);
            goto ERROR;
        }
        oacl = nacl;

        mc->op = '=';
        break;

    default:
        goto ERROR;
    }

    return oacl;

ERROR:
    return NULL;
}
#endif

/* Return a file mode change operation that sets permissions to match those
   of REF_FILE.  Return NULL (setting errno) if REF_FILE can't be accessed.  */

struct mode_change *
mode_create_from_ref (const char *ref_file)
{
  struct stat ref_stats;

  if (stat (ref_file, &ref_stats) != 0)
    return NULL;
#ifdef MY_ABC_HERE
  if (IS_FILE_SUPPORT_ACL(ref_file)) {
    SYNO_ACL *pACL = NULL;
    if ( 0 > SYNOACLGet(ref_file, -1, SYNOACL_SELF, &pACL) ) {
        return NULL;
    }
    return make_node_op_equals_acl(ref_stats.st_mode, pACL);
  }
#endif
  return make_node_op_equals (ref_stats.st_mode, CHMOD_MODE_BITS);
}

/* Return the file mode bits of OLDMODE (which is the mode of a
   directory if DIR), assuming the umask is UMASK_VALUE, adjusted as
   indicated by the list of change operations CHANGES.  If DIR, the
   type 'X' change affects the returned value even if no execute bits
   were set in OLDMODE, and set user and group ID bits are preserved
   unless CHANGES mentioned them.  If PMODE_BITS is not null, store into
   *PMODE_BITS a mask denoting file mode bits that are affected by
   CHANGES.

   The returned value and *PMODE_BITS contain only file mode bits.
   For example, they have the S_IFMT bits cleared on a standard
   Unix-like host.  */

mode_t
mode_adjust (mode_t oldmode, bool dir, mode_t umask_value,
             struct mode_change const *changes, mode_t *pmode_bits)
{
  /* The adjusted mode.  */
  mode_t newmode = oldmode & CHMOD_MODE_BITS;

  /* File mode bits that CHANGES cares about.  */
  mode_t mode_bits = 0;

  for (; changes->flag != MODE_DONE; changes++)
    {
      mode_t affected = changes->affected;
      mode_t omit_change =
        (dir ? S_ISUID | S_ISGID : 0) & ~ changes->mentioned;
      mode_t value = changes->value;

      switch (changes->flag)
        {
        case MODE_ORDINARY_CHANGE:
          break;

        case MODE_COPY_EXISTING:
          /* Isolate in 'value' the bits in 'newmode' to copy.  */
          value &= newmode;

          /* Copy the isolated bits to the other two parts.  */
          value |= ((value & (S_IRUSR | S_IRGRP | S_IROTH)
                     ? S_IRUSR | S_IRGRP | S_IROTH : 0)
                    | (value & (S_IWUSR | S_IWGRP | S_IWOTH)
                       ? S_IWUSR | S_IWGRP | S_IWOTH : 0)
                    | (value & (S_IXUSR | S_IXGRP | S_IXOTH)
                       ? S_IXUSR | S_IXGRP | S_IXOTH : 0));
          break;

        case MODE_X_IF_ANY_X:
          /* Affect the execute bits if execute bits are already set
             or if the file is a directory.  */
          if ((newmode & (S_IXUSR | S_IXGRP | S_IXOTH)) | dir)
            value |= S_IXUSR | S_IXGRP | S_IXOTH;
          break;
        }

      /* If WHO was specified, limit the change to the affected bits.
         Otherwise, apply the umask.  Either way, omit changes as
         requested.  */
      value &= (affected ? affected : ~umask_value) & ~ omit_change;

      switch (changes->op)
        {
        case '=':
          /* If WHO was specified, preserve the previous values of
             bits that are not affected by this change operation.
             Otherwise, clear all the bits.  */
          {
            mode_t preserved = (affected ? ~affected : 0) | omit_change;
            mode_bits |= CHMOD_MODE_BITS & ~preserved;
            newmode = (newmode & preserved) | value;
            break;
          }

        case '+':
          mode_bits |= value;
          newmode |= value;
          break;

        case '-':
          mode_bits |= value;
          newmode &= ~value;
          break;
        }
    }

  if (pmode_bits)
    *pmode_bits = mode_bits;
  return newmode;
}
